#coding=utf8

# This file is part of CloudControl.
#
# CloudControl is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CloudControl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with CloudControl.  If not, see <http://www.gnu.org/licenses/>.

'''
CloudControl Console related commands
'''

from cccli.command import TqlCommand
from cccli.exception import *
from cccli.printer import FakeTtySocket
from sjrpc.core.exceptions import *
import fcntl
import signal
import struct
import termios


class Command_shell(TqlCommand):
    '''Start a shell on the provided host'''

    def __init__(self, cli, argv0):
        TqlCommand.__init__(self, cli, argv0)
        self.remove_option("--direct")
        self.remove_option("--quiet")
        self.tql_filter = "&con"

    def __call__(self, argv):
        # args parse
        self.parse_args(argv)
        if len(self.args) != 1:
            raise cmdBadArgument()
        # rpccall
        ans = self.rpccall("shell", self.args[0], _status=False, _direct=True)
        # alias first object
        obj = ans['objects'][0]
        # check response is successful
        if obj['status'] != 'success':
            raise cmdError("%s: %s" % (obj["status"], obj["message"]))
        # create a callbeck function for sigwinch
        def _cb_sig_resize(*args):
            '''Callback for terminal resized'''
            size = struct.pack('HHHH', 0, 0, 0, 0)
            size = fcntl.ioctl(0, termios.TIOCGWINSZ, size)
            size = struct.unpack('HHHH', size)
            try:
                self.rpc.call('resize', None, *size)
            except RpcError as e:
                # we dont care of failure of this
                self.printer.debug("Terminal resize failed: %s" % e)
        # Connect to remote end using sjrpc
        tun = self.rpc.create_tunnel(obj["output"], endpoint=FakeTtySocket())
        try:
            signal.signal(signal.SIGWINCH, _cb_sig_resize)
            # set initial size of term
            _cb_sig_resize()
            tun.wait_shutdown()
        finally:
            # restore original signal
            signal.signal(signal.SIGWINCH, signal.SIG_DFL)
            # close tunnel if sjrpc doesn't do it
            tun.shutdown()
            # print prompt at beginning of a new line
            self.printer.out()

    def remote_functions(self):
        return set(("shell",))


class Command_console(TqlCommand):
    '''Start a vm console using libvirt'''

    def __init__(self, cli, argv0):
        TqlCommand.__init__(self, cli, argv0)
        self.remove_option("--direct")
        self.remove_option("--quiet")
        self.tql_filter = "&r=vm&status=running"

    def __call__(self, argv):
        # args parse
        self.parse_args(argv)
        if len(self.args) != 1:
            raise cmdBadArgument()
        # rpccall
        ans = self.rpccall("console", self.args[0], _status=False, _direct=True)
        # alias first object
        obj = ans['objects'][0]
        # check everything is ok
        if obj['status'] != 'success':
            raise cmdError("%s: %s" % (obj["status"], obj["message"]))
        # Connect to remote end using sjrpc
        tun = self.rpc.create_tunnel(obj["output"], endpoint=FakeTtySocket())
        try:
            tun.wait_shutdown()
        finally:
            # close tunnel if sjrpc doesn't do it
            tun.shutdown()
            # print prompt at beginning of a new line
            self.printer.out()

    def remote_functions(self):
        return set(("console",))
